/*
  Questo modulo permette di recuperare i file appartenenti al filesystem TCFS
  TCFS version 3.0

  Copyright (C) 2000, 2001 Giuseppe Cupo
*/

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "alloc.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "md5.h"
#include "tcfsLinuxHeader.h"

extern module tcfs_module;

//PLUGIN HEADER
typedef struct _plugin_header{
  char *server_mime;      //Mime type determinato dal server
  int tcfsVersion;        //versione di TCFS  
  int cipherID;           //ID del cifratore
} plugin_header;


int  tcfs_check_hash(const char*, const char*, char*, int);
int send_plugin_header(plugin_header *, request_rec *);
int pluginHeaderSize(plugin_header *);
/*
  Handler della richiesta.
*/
static int TcfsHandler(request_rec *r){
  int error;
  int n, i;
  FILE *f;
  char buffer[TCFS_HBLOCK_SIZE];

  //struttura contenente l'header aggiuntivo da inviare al Navigator
  plugin_header pluginHeader;

  //  ap_log_reason("\"tcfs_module in action\"\n", r->filename, r);

  // Controlliamo che la richiesta sia di tipo GET
  if(r->method_number != M_GET) return DECLINED;

  
  // Controlliamo che il file richiesto esista  
  if(r->finfo.st_mode == 0) return NOT_FOUND;


  // ap_log_reason("Il file richiesto esiste\n", r->filename, r);
  //apertura file
  f = ap_pfopen(r->pool, r->filename, "r");

  if(f == NULL){
    ap_log_reason("Filename permission deny server access\n", r->filename, r);
    return FORBIDDEN;
  }

  /*
    Adesso che il file e' aperto dobbiamo verificare se e' cifrato o meno: 
    i primi 4 byte devono coincidere con "TCFS".
    Questo e' l'unico controllo che possiamo fare con questa versione di TCFS
  */
  
  //controlliamo che la dimensione del file multipla di TCFS_HBLOCK_SIZE
  if(!(r->finfo.st_size%TCFS_HBLOCK_SIZE));
  else{
    ap_log_reason("TCFS file size control failed: NO tcfs file...!\n", r->filename, r);
    ap_pfclose(r->pool, f);    
    return DECLINED;
  }

  //verifica dell'hashing: leggiamo l'header TCFS: il primo blocco di TCFS_HBLOCK_SIZE
  n = fread(buffer, sizeof(char), TCFS_HBLOCK_SIZE, f);  

  if(n == TCFS_HBLOCK_SIZE){
    error=tcfs_check_hash(buffer, NULL, buffer+TCFS_HEADER_SIZE, TCFS_HEADER_SIZE);
    if(error<0){
      ap_log_reason("TCFS check hash cotrol failed: NO tcfs file...!\n", r->filename, r);
      ap_pfclose(r->pool, f);
      return DECLINED;
    }    
  }else{
    ap_pfclose(r->pool, f);
    return DECLINED;
  }


  //controlliamo che i primi 4 byte siamo "TCFS"
  if(strncmp(buffer, "TCFS", 4)){
    ap_pfclose(r->pool, f);
    return DECLINED;
  }


  //memorizziamo il MIME type determinato dal server 
  //Se il server non riconosce il content type, consideriamo plain/text per default
  if(r->content_type) 
    pluginHeader.server_mime = ap_pstrcat(r->pool, r->content_type, " ", NULL);
  else 
    pluginHeader.server_mime = ap_pstrcat(r->pool, "text/plain", " ", NULL);

  //Aggiorniamo il MIME type dell'header HTTP
  r->content_type = ap_pstrcat(r->pool, "application/x-tcfs", NULL);

  
  //Aggiorniamo la taglia della risposta
  ap_set_content_length(r, r->finfo.st_size + pluginHeaderSize(&pluginHeader));
  
  //Inviamo al client il tempo di ultima modifica della risorsa
  ap_update_mtime(r, r->finfo.st_mtime);
  ap_set_last_modified(r); 


  //settiamo il timeout
  ap_hard_timeout("mod_tcfs", r);

  //invio header HTTP
  ap_send_http_header(r);
  

  //inseriamo nel PLUGIN HEADER la versione di TCFS e l'Id del cifratore
  pluginHeader.tcfsVersion = 3; 
  pluginHeader.cipherID = 0;    

  //i dati vengono inviati solo se la richiesta non e' header only  
  if(!r->header_only){ 

    //invio PLUGIN HEADER
    send_plugin_header(&pluginHeader, r);

    //invio il blocco gia' letto dal file
    ap_rwrite(buffer, TCFS_HBLOCK_SIZE, r);


    //invio il resto del file 
    ap_send_fd(f, r);
  }
  
    
  ap_pfclose(r->pool, f);
    
  ap_kill_timeout(r);

  return OK;

}// end handler


/*
  Associamo una stringa di handler all'handler della richiesta
*/
static handler_rec aHandlers[] = {
  {"tcfs_handler", TcfsHandler},
  {NULL}
};


module tcfs_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
    NULL,			/* create per-directory config structure */
    NULL,			/* merge per-directory config structures */
    NULL,			/* create per-server config structure */
    NULL,			/* merge per-server config structures */
    NULL,			/* command table */
    aHandlers,	        	/* handlers */
    NULL,			/* translate_handler */
    NULL,			/* check_user_id */
    NULL,			/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* pre-run fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};



int pluginHeaderSize(plugin_header *ph)
{
  return (2+strlen(ph->server_mime));
}

int send_plugin_header(plugin_header *ph, request_rec *r)
{
  char info[2];

  // byte 1 = Versione tcfs
  // byte 2 = cipher ID  
  info[0] = ph->tcfsVersion;
  info[1] = ph->cipherID;  
 
  //invio mime determinato dal Server
  ap_rwrite(ph->server_mime, strlen(ph->server_mime), r);

  //invio informazioni per il plugin 
  ap_rwrite(info, 2, r); 

  return 0;
}


//definizione delle funzioni per la verifica dell'hashing
/*
 * Calcola l'hashing di un blocco
*/
int tcfs_hash_block(const char *buffer, int size, char *out)
{
	MD5_CTX context;
	unsigned char digest[TCFS_HASH_SIZE];

	memset (digest, 0, TCFS_HASH_SIZE);

	MD5Init (&context);
	MD5Update (&context, (unsigned char*)buffer, (unsigned int)size);
	MD5Final (digest, &context);

        memcpy (out, digest, TCFS_HASH_SIZE);
                
        return 0;
}


int tcfs_check_hash(const char *buffer, const char *bk, char *hash, int size)
{ 
	char newhash[TCFS_HASH_SIZE];
	char buf[size+KEYSIZE];
   
	memcpy (buf, buffer, size);

	if(bk){
	  memcpy (buf+size, bk, KEYSIZE);
	  tcfs_hash_block (buf, size+KEYSIZE, newhash);
	} else {
	  tcfs_hash_block (buf, size, newhash);
	}

	if (!memcmp (hash, newhash, TCFS_HASH_SIZE))
	   return 0;
        else 
	   return -1;
}               


